r"""Utility functions and classes for building applications.

Currently, these utilities are primarily intended for use in example applications and tutorials
of this library. It is not recommended to build external applications on these while it is in a
non-stable release version (<1.0).

.. note::
    These auxiliary modules should only depend on the :mod:`.core` library and third-party packages.

"""
