r"""Utility functions for `SimpleITK <https://simpleitk.org/>`_ data objects."""

from .imageio import read_image, write_image


# fmt: off
__all__ = (
    "read_image",
    "write_image",
)
# fmt: on
